import redis
import uuid
import datetime
import time
import random
from python_grains.utils import RedisClientWithRetry

DEFAULT_RETRY_TIMES = 0
DEFAULT_RETRY_DELAY = 100
LOCK_TTL = 100000 # in milliseconds
CLOCK_DRIFT_FACTOR = 0.01
USER_QUEUE_TTL = 3600 # in seconds
MAX_PROCESSING_RETRIES = 3

RELEASE_LUA = '''
if redis.call("GET", KEYS[1]) == ARGV[1] then
    return redis.call("DEL", KEYS[1])
else
    return 0
end'''

class LockError(Exception): pass

class Lock(object):

    def __init__(self, idx, conn, ttl,  # ttl in milliseconds
                 retry_times=DEFAULT_RETRY_TIMES,
                 retry_delay=DEFAULT_RETRY_DELAY):

        self.type = type
        self.lock_id = 'lck:' + idx
        self.retry_times = retry_times
        self.retry_delay = retry_delay
        self.ttl = ttl

        if not isinstance(conn, (redis.StrictRedis, RedisClientWithRetry)):
            raise LockError('No valid connection details provided')

        self.redis_node = conn
        self.add_release_script()

    def add_release_script(self):
        if not hasattr(self.redis_node, 'release_lock'):
            self.redis_node.release_lock = self.redis_node.register_script(RELEASE_LUA)

    def __enter__(self):
        acquired, validity = self.acquire_with_validity()
        if not acquired:
            raise LockError('Failed to acquire lock.')
        return validity

    def __exit__(self, exc_type, exc_value, traceback):
        self.release()

    def release(self):
        self.release_node(self.redis_node)

    def _total_ms(self, delta):
        """
        Get the total number of milliseconds in a timedelta object with
        microsecond precision.
        """
        delta_seconds = delta.seconds + delta.days * 24 * 3600
        return (delta.microseconds + delta_seconds * 10**6) / 10**3

    def acquire_node(self, node):
        """
        acquire a single redis node
        """
        try:
            return node.set(self.lock_id, self.lock_key, nx=True, px=self.ttl)
        except (redis.exceptions.ConnectionError, redis.exceptions.TimeoutError):
            return False

    def release_node(self, node):
        """
        release a single redis node
        """
        # use the lua script to release the lock in a safe way
        try:
            node.release_lock(keys=[self.lock_id], args=[self.lock_key])
        except (redis.exceptions.ConnectionError, redis.exceptions.TimeoutError):
            pass

    def locked(self):

        if self.redis_node.get(self.lock_id):
            return True
        return False

    def acquire(self):
        acquired, validity = self._acquire()
        return acquired

    def acquire_with_validity(self):
        return self._acquire()

    def _acquire(self):

        # lock_key should be random and unique
        self.lock_key = uuid.uuid4().hex

        for retry in range(self.retry_times + 1):
            start_time = datetime.datetime.utcnow()

            # acquire the lock
            acquired = self.acquire_node(self.redis_node)

            end_time = datetime.datetime.utcnow()
            elapsed_milliseconds = self._total_ms(end_time - start_time)

            # Add 2 milliseconds to the drift to account for Redis expires
            # precision, which is 1 millisecond, plus 1 millisecond min drift
            # for small TTLs.
            drift = (self.ttl * CLOCK_DRIFT_FACTOR) + 2

            validity = self.ttl - (elapsed_milliseconds + drift)

            if validity > 0 and acquired:
                return True, validity
            else:
                self.release_node(self.redis_node)
                time.sleep(random.randint(0, self.retry_delay) / 1000)

        return False, 0


class LuaScripts(object):

    @classmethod
    def get_latest_data(cls):
        '''
        KEYS[1]: current version key
        ARGV[1]: data key prefix
        ARGV[2]: data ttl
        '''

        return '''
            local current_version = redis.call('GET', KEYS[1])
            if current_version == false then
                return nil
            end
            local data_key = ARGV[1] .. current_version
            local data_string = redis.call('GET', data_key)
            redis.call('EXPIRE', data_key, tonumber(ARGV[2]))
            
            return data_string
            '''

    @classmethod
    def set_data_with_version(cls):

        '''
        KEYS[1]: current version key
        KEYS[2]: data key
        ARGV[1]: new version number
        ARGV[2]: data string
        ARGV[3]: data prefix
        ARGV[4]: data ttl
        '''

        return '''
            local result = {}
            local current_cache_version = redis.call('GET', KEYS[1])
            if current_cache_version == false or tonumber(current_cache_version) < tonumber(ARGV[1]) then
                redis.call('SET', KEYS[2], ARGV[2])
                redis.call('SET', KEYS[1], ARGV[1])
                redis.call('EXPIRE', KEYS[2], tonumber(ARGV[4]))
                table.insert(result, '1')
                table.insert(result, nil)
            elseif tonumber(current_cache_version) ==  tonumber(ARGV[1]) then
                table.insert(result, '1')
                table.insert(result, nil)
            else
                table.insert(result, '0')
                table.insert(result, redis.call('GET', ARGV[3] .. current_cache_version))
            end
            return result
        '''

    @classmethod
    def set_download_in_progress(cls):

        '''
        KEYS[1]: download in progress key
        ARGV[1]: download in progress ttl
        '''

        return '''
            return redis.call('SET', KEYS[1], '1', 'NX', 'EX', tonumber(ARGV[1]))
        '''


from requests.sessions import Session
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry
from requests.auth import HTTPBasicAuth

from python_grains.utils import search_obj_from_array

DEFAULT_PATH = '/pseudonyms/'

class AliassingClient(object):

    alias_postfix = 'Alias'

    def __init__(self,
                 hostname,
                 username,
                 password,
                 scheme='https',
                 path=None,
                 retry=3,
                 timeout=5):

        self.hostname = hostname
        self.scheme = scheme
        self.path = path or DEFAULT_PATH
        self.username = username
        self.password = password
        self.default_retry = retry
        self.default_timeout = timeout

    def get_from_service_external(self,
                                  by_alias,
                                  by_input,
                                  timeout=None,
                                  create=True):

        result = {}

        if by_alias:

            r = self.http_session.post(url=self.endpoint,
                                       json=list(by_alias.values()),
                                       auth=self.auth,
                                       params={'create': False},
                                       timeout=timeout or self.default_timeout)

            response = r.json()

            for k,a in by_alias.items():
                _obj = search_obj_from_array(response, 'alias', a['alias'])
                if _obj:
                    if _obj['status'] in (200, 201):
                        result[k] = {**_obj, 'status': 200}
                    else:
                        result[k] = _obj
                else:
                    result[k] = {**a, 'status': 404}

        if by_input:

            r = self.http_session.post(url=self.endpoint,
                                       json=list(by_input.values()),
                                       auth=self.auth,
                                       params={'create': create},
                                       timeout=timeout or self.default_timeout)
            t2 = r.elapsed.total_seconds()
            response = r.json()
            for k, a in by_input.items():
                _obj = search_obj_from_array(response, 'input', a['input'])
                if _obj:
                    if _obj['status'] in (200, 201):
                        result[k] = {**_obj, 'status': 200}
                    else:
                        result[k] = _obj
                else:
                    result[k] = {**a, 'status': 404}

        return result

    def get(self,
            input,
            timeout=None,
            create=False):

        if not input:
            return []
        else:

            to_service = {}
            result = {}
            input = dict(enumerate(input))
            for k, a in input.items():
                to_service[k] = a

            if to_service:
                by_input = {k: r for k, r in to_service.items() if 'input' in r}
                by_alias = {k: r for k, r in to_service.items() if not 'input' in r}

                result = self.get_from_service_external(by_alias=by_alias, by_input=by_input, timeout=timeout, create=create)

        final_result = [None] * (max(input) + 1)
        for k, a in input.items():
            final_result[k] = {k:v for k,v in result.get(k, {**a, 'status': 404}).items() if k in ('input', 'status', 'variable', 'alias')}

        return final_result

    def preprocess_input(self,
                         input):

        return [{'variable': k, 'input': v} for k,v in input.items()]

    def postprocess_output(self,
                           output):

        return {a['variable'] + self.alias_postfix: a['alias'] for a in output if a['status'] in (200, 201)}

    @property
    def auth(self):

        return HTTPBasicAuth(self.username, self.password)

    @property
    def http_session(self):

        if not hasattr(self, '_http_session') or self._http_session is None:
            self._http_session = Session()
            retry_strategy = Retry(
                total=3,
                backoff_factor=1,
                status_forcelist=[429, 500, 502, 503, 504],
                allowed_methods=['GET', 'POST'],
                respect_retry_after_header=True
            )
            self._http_session.mount(self.host, adapter=HTTPAdapter(max_retries=retry_strategy))
        return self._http_session

    @property
    def host(self):
        return f'{self.scheme}://{self.hostname}'

    @property
    def endpoint(self):

        return f'{self.host}{self.path}'



import redis
from redis.retry import Retry
from redis.backoff import DecorrelatedJitterBackoff
import socket
import time
import datetime
import pytz
import logging

class RedisClientWithRetry(object):

    @classmethod
    def from_config(cls, *args, max_retry=5, **kwargs):
        backoff = DecorrelatedJitterBackoff(cap=2.0, base=0.5)
        retry = Retry(retries=max_retry, backoff=backoff)
        redis_client = redis.Redis(*args,
                                   retry_on_error=[redis.exceptions.ConnectionError,
                                                   redis.exceptions.TimeoutError,
                                                   socket.timeout],
                                   retry=retry,
                                   **kwargs)
        return redis_client

def chunker(seq, size):
    return (seq[pos:pos + size] for pos in range(0, len(seq), size))

def search_obj_from_array(array, fieldname, fieldvalue):
    _obj = [o for o in array if o.get(fieldname, str(fieldvalue) + 'not_it') == fieldvalue]
    if len(_obj) == 0:
        return None
    return _obj[0]


class Timer(object):

    def __init__(self, name=None, info=None, logger=None, logger_name='timings', sub=False):

        self.start = time.time()
        self.end = None
        self.timings = []
        self.info = info
        self.logger = logger or logging.getLogger(logger_name)
        self.sub = sub
        self.name = name or 'timer'

    def subtimer(self, name, info=None):
        t = Timer(name=name, info=info, logger=self.logger, sub=True)
        self.timings.append(t)
        return t


    def register(self, name, info=None):

        assert self.end is None, 'This timer instance is already done'
        assert name != 'total', 'Name cannot be `total`'

        self.timings.append((time.time(), name, info))

    def done(self, send_log=True, loglvl='info', logverbose=True):

        self.end = time.time()
        self.timings = sorted(self.timings, key=lambda x: x.start if isinstance(x, Timer) else x[0])

        if send_log and not self.logger is None:
            getattr(self.logger, loglvl)('Timing done', extra=self.result(verbose=logverbose))

    def result(self, verbose=True):

        if self.end is None:
            self.done(send_log=False)

        durations = [x[1]-x[0] for x in zip([self.start, *self.timings[:-1]], self.timings)]

        if verbose:

            timings = [{
                'name': x[1],
                'time': pytz.utc.localize(datetime.datetime.utcfromtimestamp(x[0])).isoformat(),
                'info': x[2],
                'duration': duration
            } for x, duration in zip(self.timings, durations)]

            r = {
                'start': pytz.utc.localize(datetime.datetime.utcfromtimestamp(self.start)).isoformat(),
                'end': pytz.utc.localize(datetime.datetime.utcfromtimestamp(self.end)).isoformat(),
                'duration': self.total,
                'info': self.info,
                'timings': timings}

        else:

            r = {
                x[1]: duration
                for x, duration in zip(self.timings, durations)}
            r.update({'total': self.total})

        return r

    @property
    def total(self):
        return self.end - self.start





import os

class LuaScripts(object):

    @classmethod
    def set_profile_if_new_user(cls,
                                user_ttl,
                                user_ttl_margin,
                                profile_ttl,
                                profile_ttl_margin,
                                new_user_detection_ttl):
        '''
        KEYS[1]: user key
        KEYS[2]: the key of the set with all profile keys
        KEYS[3]: the key for the profile
        KEYS[4]: the key for the profile key pool
        KEYS[5]: the key for new user detection
        ARGV[1]: profile update info
        ARGV[2]: current timestamp
        '''

        return '''
            local new_user = (redis.call('EXISTS', KEYS[1]) == 0)
            if not new_user then
                return "0"
            end
            local profile_key = KEYS[3]
            local profile_pool_key = KEYS[4]
            local timestamp_now = tonumber(ARGV[2])
            redis.call('SET', KEYS[1], profile_key)
            redis.call('SADD', KEYS[4], profile_key)

            local profile_data = cjson.decode(ARGV[1])

            redis.call('HMSET', profile_key, unpack(profile_data))

            -- add profile_key to sorted set of all profile keys
            redis.call('ZADD', KEYS[2], timestamp_now, profile_key)            

            -- set key for new user detection
            redis.call('SET', KEYS[5], "1", 'EX', {new_user_detection_ttl}, 'NX')

            -- set expirations
            redis.call('EXPIRE', KEYS[1], {user_ttl})          
            redis.call('EXPIRE', profile_key, {profile_ttl})                
            redis.call('EXPIRE', profile_pool_key, {user_ttl_with_margin})  
            return "1"    
        '''.format(user_ttl=user_ttl,
                   user_ttl_with_margin=user_ttl + user_ttl_margin,
                   profile_ttl=profile_ttl + profile_ttl_margin,
                   new_user_detection_ttl=new_user_detection_ttl)

    @classmethod
    def update_profile(cls,
                       user_ttl,
                       user_ttl_margin,
                       profile_ttl,
                       profile_ttl_margin,
                       value_postfix,
                       created_postfix,
                       updated_postfix,
                       expire_postfix,
                       new_user_detection_ttl):
        '''
        KEYS[1]: user key containing profile key
        KEYS[2]: the key of the set with all profile keys
        KEYS[3]: the key of the set with all pointers
        KEYS[4]: the key for the profile (ARGV[1])
        KEYS[5]: the key for the profile id pool (ARGV[2])
        KEYS[6]: the key for new user detection
        KEYS[7]: the key for the stitch pool
        ARGV[1]: profile update info (ARGV[3])
        ARGV[2]: current timestamp (ARGV[4])
        ARGV[3]: session id (ARGV[5])
        ARGV[4]: user id (ARGV[6])
        ARGV[5]: array with pointers (can be empty) (ARGV[7])
        ARGV[6]: array with stitch keys (can be empty)
        '''
        return '''
            local new_user = (redis.call('EXISTS', KEYS[1]) == 0)
            local new_user_detected = "1"
            local profile_key = KEYS[4]
            local profile_pool_key = KEYS[5]
            local new_profile = false
            local timestamp_now = tonumber(ARGV[2])
            if new_user then
                redis.call('SET', KEYS[1], profile_key)
                -- set the key for detecting a new user
                redis.call('SET', KEYS[6], "1", 'EX', {new_user_detection_ttl}, 'NX')
                -- create profile pool
                redis.call('SADD', profile_pool_key, profile_key)

                new_profile = true
            else
                profile_key = redis.call('GET', KEYS[1])

                -- profile has been offloaded or expired
                if profile_key == "none" or (redis.call('EXISTS', profile_key)==0) then
                   profile_key = KEYS[4]
                   redis.call('SET', KEYS[1], profile_key)
                   redis.call('SADD', profile_pool_key, profile_key) 
                   new_profile = true
                end

                if redis.call('GET', KEYS[6]) == "1" then
                    new_user_detected = "1"
                else
                    new_user_detected = "0"
                end

            end

            local user_pointers = cjson.decode(ARGV[5])                          -- set user pointers
            for idx, el in pairs(user_pointers) do
                redis.call('SET', el, ARGV[4])
                redis.call('SADD', KEYS[3], el)
            end
            redis.call('EXPIRE', KEYS[3], {user_ttl_with_margin})

            local all_user_pointers = redis.call('SMEMBERS', KEYS[3])  -- set expiration on user pointers
            for idx, el in pairs(all_user_pointers) do
                local pointed_user_id = redis.call('GET', el)
                if pointed_user_id == nil or pointed_user_id~=ARGV[4] then
                    redis.call('SREM', KEYS[3], el)
                else
                    redis.call('EXPIRE', el, {user_ttl_with_margin})
                end
            end

            -- add stitch keys
            local stitch_keys = cjson.decode(ARGV[6])
            for idx, el in pairs(stitch_keys) do
                redis.call('ZADD', el, timestamp_now, ARGV[4])
                redis.call('EXPIRE', el, {user_ttl_with_margin})
                redis.call('ZADD', KEYS[7], timestamp_now, el)
                redis.call('EXPIRE', KEYS[7], {user_ttl_with_margin})
            end

            -- add user_hash_key. created and updated to profile
            redis.call('HMSET', profile_key, '_user_hash', KEYS[1], '_updated', timestamp_now)
            if new_profile then
                redis.call('HSET', profile_key, '_created', timestamp_now)
            end

            local function update_profile_prop (p_key, v_key, obj, field_type)
                local current_value = redis.call('HGET', p_key, v_key)
                local v = obj["value"]
                if field_type == 'counter' then
                    redis.call('HINCRBY', p_key, v_key, v)
                elseif field_type == 'sfloat' then
                    if current_value == false or tonumber(current_value) > tonumber(v) then
                        redis.call('HSET', p_key, v_key, v)
                    end 
                elseif field_type == 'lfloat' then
                    if current_value == false or tonumber(current_value) < tonumber(v) then
                        redis.call('HSET', p_key, v_key, v)
                    end
                elseif field_type == 'sumfloat' then
                    redis.call('HINCRBYFLOAT', p_key, v_key, v)
                elseif field_type == 'lustring' then
                     redis.call('HSET', p_key, v_key, v)
                elseif field_type == 'lufloat' then
                     redis.call('HSET', p_key, v_key, v)
                elseif field_type == 'fcstring' then
                    if current_value == false then
                        redis.call('HSET', p_key, v_key, v)
                    end
                elseif field_type == 'orbool' then
                    if (current_value == false or current_value == '1') or v == true then
                        redis.call('HSET', p_key, v_key, '1')
                    else
                        redis.call('HSET', p_key, v_key, '0')
                    end
                elseif field_type == 'andbool' then
                    if (current_value == false or current_value == '1') and v == true then
                        redis.call('HSET', p_key, v_key, '1')
                    else
                        redis.call('HSET', p_key, v_key, '0')
                    end
                elseif field_type == 'fcfloat' then
                    if current_value == false then
                        redis.call('HSET', p_key, v_key, v)
                    end    
                elseif field_type == 'tw30count' then
                    local threshold_ts = tonumber(obj["thr"])
                    local new_counter_data = {{}}
                    for day_key, val in pairs(cjson.decode(v)) do
                        if tonumber(day_key) >= threshold_ts then
                            new_counter_data[tostring(day_key)] = val
                        end
                    end
                    if current_value == false then
                        local counter_data_str = cjson.encode(new_counter_data)
                        redis.call('HSET', p_key, v_key, counter_data_str)
                    else
                        local counter_data = cjson.decode(current_value)
                        for k, val in pairs(counter_data) do
                            if tonumber(k) >= threshold_ts then
                                new_counter_data[k] = tonumber((new_counter_data[k] or 0)) + val
                            end
                        end
                        local counter_data_str = cjson.encode(new_counter_data)
                        redis.call('HSET', p_key, v_key, counter_data_str)
                    end      
                end 
            end

            local function update_profile(dataset, field_type)
                -- function to update the data keys
                for idx, obj in pairs(dataset) do

                    local value_key = obj["key"] .. "{value_postfix}"
                    local updated_key = obj["key"] .. "{updated_postfix}"
                    local created_key = obj["key"] .. "{created_postfix}"
                    local expire_key = obj["key"] .. "{expire_postfix}"

                    if redis.call('HEXISTS', profile_key, value_key) == 0 then
                        -- the key does not exist yet
                        if (obj.expire_at == nil or obj.expire_at > timestamp_now) and obj["value"] ~= nil then
                            -- only create if the new expire time is in the future
                            update_profile_prop(profile_key, value_key, obj, field_type)
                            redis.call('HMSET', profile_key, updated_key, timestamp_now, created_key, timestamp_now)
                            if obj.expire_at ~= nil then
                                redis.call('HSET', profile_key, expire_key, obj.expire_at)  
                            end
                        end
                    else
                        -- the key already exists
                        local deleted_key = false
                        local current_expire = tonumber(redis.call('HGET', profile_key, expire_key))
                        if (current_expire ~= nil and current_expire <= timestamp_now) or (obj.expire_at ~= nil and obj.expire_at <= timestamp_now) then
                            -- first delete if the old or new expire time is in the past
                            redis.call('HDEL', profile_key, value_key)
                            redis.call('HDEL', profile_key, updated_key)
                            redis.call('HDEL', profile_key, created_key)
                            redis.call('HDEL', profile_key, expire_key)
                            deleted_key = true
                        end
                        if obj.expire_at == nil or obj.expire_at > timestamp_now then
                            -- create/update the key if its not expired
                            if obj["value"] ~= nil or not deleted_key then
                                update_profile_prop(profile_key, value_key, obj, field_type)
                                redis.call('HSET', profile_key, updated_key, timestamp_now)  
                                if deleted_key then
                                    redis.call('HSET', profile_key, created_key, timestamp_now)
                                end
                                if obj.expire_at ~= nil then
                                    redis.call('HSET', profile_key, expire_key, obj.expire_at)
                                end
                            end  
                        end
                    end                        
                end
            end

            -- do the actual profile update
            local profile_update = cjson.decode(ARGV[1])

            if profile_update.counters ~= nil then
                update_profile(profile_update.counters, 'counter')
            end

            if profile_update.smallest_floats ~= nil then
                update_profile(profile_update.smallest_floats, 'sfloat')
            end

            if profile_update.largest_floats ~= nil then
                update_profile(profile_update.largest_floats, 'lfloat')
            end

            if profile_update.summed_floats ~= nil then
                update_profile(profile_update.summed_floats, 'sumfloat')
            end

            if profile_update.last_updated_strings ~= nil then
                update_profile(profile_update.last_updated_strings, 'lustring')
            end

            if profile_update.first_created_strings ~= nil then
                update_profile(profile_update.first_created_strings, 'fcstring')
            end

            if profile_update.first_created_floats ~= nil then
                update_profile(profile_update.first_created_floats, 'fcfloat')
            end

            if profile_update.last_updated_floats ~= nil then
                update_profile(profile_update.last_updated_floats, 'lufloat')
            end

            if profile_update.and_booleans ~= nil then
                update_profile(profile_update.and_booleans, 'andbool')
            end

            if profile_update.or_booleans ~= nil then
                update_profile(profile_update.or_booleans, 'orbool')
            end
            
            if profile_update.time_window30_counters ~= nil then
                update_profile(profile_update.time_window30_counters, 'tw30count')
            end

            -- add profile_key to sorted set of all profile keys
            redis.call('ZADD', KEYS[2], timestamp_now, profile_key)            

            -- set expirations
            redis.call('EXPIRE', KEYS[1], {user_ttl})          
            redis.call('EXPIRE', profile_key, {profile_ttl})                
            redis.call('EXPIRE', profile_pool_key, {user_ttl_with_margin})    

            return new_user_detected             
        '''.format(user_ttl=user_ttl,
                   user_ttl_with_margin=user_ttl + user_ttl_margin,
                   profile_ttl=profile_ttl + profile_ttl_margin,
                   updated_postfix=updated_postfix,
                   value_postfix=value_postfix,
                   created_postfix=created_postfix,
                   expire_postfix=expire_postfix,
                   new_user_detection_ttl=new_user_detection_ttl)

    @classmethod
    def get_user_profile(cls,
                         updated_postfix,
                         value_postfix,
                         created_postfix,
                         expire_postfix):
        '''
        KEYS[1]: profile_pool_key
        ARGV[1]: the current time
        '''

        return '''
            local profiles = {{}}
            local profile_keys = redis.call('SMEMBERS', KEYS[1])
            for idx, profile_key in pairs(profile_keys) do
                table.insert(profiles, profile_key)
                -- delete expired keys
                local keys = redis.call('HKEYS', profile_key)
                for i, key in pairs(keys) do
                    if string.len(key) > 2 and string.sub(key, -2, -1) == '{expire_postfix}' then
                        local postfix = string.sub(key, -2, -1)
                        local pre = string.sub(key, 1, -3)
                        if tonumber(redis.call('HGET', profile_key, key)) < tonumber(ARGV[1]) then
                            redis.call('HDEL', profile_key, key)
                            redis.call('HDEL', profile_key, pre .. '{value_postfix}')
                            redis.call('HDEL', profile_key, pre .. '{created_postfix}')
                            redis.call('HDEL', profile_key, pre .. '{updated_postfix}')
                        end
                    end
                end                    
                table.insert(profiles, redis.call('HGETALL', profile_key))
            end
            return profiles               
        '''.format(updated_postfix=updated_postfix,
                   value_postfix=value_postfix,
                   created_postfix=created_postfix,
                   expire_postfix=expire_postfix)

    @classmethod
    def get_stitched_user_profile(cls,
                                  updated_postfix,
                                  value_postfix,
                                  created_postfix,
                                  expire_postfix):
        '''
        ARGV[1]: json array of profile_pool_keys
        ARGV[2]: the current time
        '''

        return '''
            local result = {{}}
            local profile_pool_keys = cjson.decode(ARGV[1])

            for idx, profile_pool_key in pairs(profile_pool_keys) do
                local profiles = {{}}
                local profile_keys = redis.call('SMEMBERS', profile_pool_key)

                for ix, profile_key in pairs(profile_keys) do
                    table.insert(profiles, profile_key)
                    -- delete expired keys
                    local keys = redis.call('HKEYS', profile_key)
                    for i, key in pairs(keys) do
                        if string.len(key) > 2 and string.sub(key, -2, -1) == '{expire_postfix}' then
                            local postfix = string.sub(key, -2, -1)
                            local pre = string.sub(key, 1, -3)
                            if tonumber(redis.call('HGET', profile_key, key)) < tonumber(ARGV[1]) then
                                redis.call('HDEL', profile_key, key)
                                redis.call('HDEL', profile_key, pre .. '{value_postfix}')
                                redis.call('HDEL', profile_key, pre .. '{created_postfix}')
                                redis.call('HDEL', profile_key, pre .. '{updated_postfix}')
                            end
                        end
                    end                    
                    table.insert(profiles, redis.call('HGETALL', profile_key))
                end
                table.insert(result, profile_pool_key)
                table.insert(result, profiles)
            end

            return result               
        '''.format(updated_postfix=updated_postfix,
                   value_postfix=value_postfix,
                   created_postfix=created_postfix,
                   expire_postfix=expire_postfix)

    @classmethod
    def completely_remove_from_cache(cls):
        '''
        KEYS[1]: profile_key
        KEYS[2]: profile pool key
        KEYS[3]: all profile keys key
        '''

        return '''
                redis.call('ZREM', KEYS[3], KEYS[1])
                redis.call('SREM', KEYS[2], KEYS[1])
                redis.call('DEL', KEYS[1])
            '''

    @classmethod
    def get_all_user_pointers(cls,
                              user_ttl,
                              user_ttl_margin):
        '''
        KEYS[1]: user pointer set key
        ARGV[1]: user id
        ARVG[2]: json string of user pointer from database
        '''

        return '''        
            local all_user_pointers = redis.call('SMEMBERS', KEYS[1])
            local user_pointers_from_db = cjson.decode(ARGV[2])

            local result = {{}}
            for idx, el in pairs(all_user_pointers) do
                local pointed_user_id = redis.call('GET', el)
                if pointed_user_id == nil or pointed_user_id~=ARGV[1] then
                    redis.call('SREM', KEYS[1], el)
                else
                    table.insert(result, el)
                end
            end
            for idx, el in pairs(user_pointers_from_db) do
                if redis.call('SISMEMBER', KEYS[1], el) == 0 then
                    local pointed_user_id = redis.call('GET', el)
                    if pointed_user_id == nil or pointed_user_id == ARGV[1] then
                        if pointed_user_id == ARGV[1] then
                            redis.call('SET', el, ARGV[1], 'EX', {user_ttl_with_margin})
                        end
                        redis.call('SADD', KEYS[1], el) 
                        table.insert(result, el)
                    end
                end
            end
            redis.call('EXPIRE', KEYS[1], {user_ttl_with_margin})

            return result
        '''.format(user_ttl_with_margin=user_ttl + user_ttl_margin)

    @classmethod
    def get_multiple_stitches(cls):
        '''
        ARGV[1]: json string of stitch keys
        ARVG[2]: recently checked db postfix
        '''

        return '''        
                local stitch_keys = cjson.decode(ARGV[1])

                local result = {}

                for idx, el in pairs(stitch_keys) do

                    local stitch_key_result = {}
                    local user_ids = redis.call('ZRANGE', el, '-inf', '+inf')
                    table.insert(stitch_key_result, 'user_ids')
                    table.insert(stitch_key_result, user_ids)

                    recent_check = redis.call('GET', el .. ARGV[2])
                    table.insert(stitch_key_result, 'recent_check')
                    table.insert(stitch_key_result, recent_check)

                    table.insert(result, el)
                    table.insert(result, stitch_key_result)

                end

                return result
            '''

    @classmethod
    def switch_profiles(cls,
                        user_ttl,
                        user_ttl_margin):
        '''
        Make sure you place the new profile in the database first
        KEYS[1]: profile pool key
        KEYS[2]: all profile keys key
        ARGV[1]: json array of old profile_keys
        ARGV[2]: new_profile_key
        '''

        return '''
            local old_profile_keys = cjson.decode(ARGV[1])
            redis.call('ZREM', KEYS[2], unpack(old_profile_keys))
            redis.call('DEL', unpack(old_profile_keys))

            local pool_exists = (redis.call('EXISTS', KEYS[1]) == 1)

            if pool_exists then 
                -- start by adding the new value, otherwise ttl might go wrong when removing causes the set to be empty (i.e. it removes it)
                -- note that the new profile key needs to be different from all old profile keys
                redis.call('SADD', KEYS[1], ARGV[2])
                redis.call('SREM', KEYS[1], unpack(old_profile_keys))
            else
                redis.call('SADD', KEYS[1], ARGV[2])
                redis.call('EXPIRE', KEYS[1], {user_ttl_with_margin})
            end
        '''.format(user_ttl_with_margin=user_ttl + user_ttl_margin)

    @classmethod
    def finalize_profile(cls, local=False):
        '''
        KEYS[1]: profile_key
        '''

        return '''
            local user_hash_key = redis.call('HGET', KEYS[1], '_user_hash')

            if user_hash_key == false or user_hash_key == nil then
                return nil
            end

            local current_active_profile = "0"

            if redis.call('GET', user_hash_key) == KEYS[1] then

                if "{local}" == "1" then
                    redis.call('SET', user_hash_key, "none", 'EX', 110)
                else
                    redis.call('SET', user_hash_key, "none", 'KEEPTTL')
                end
                current_active_profile = "1"
            end

            local profile = {{}}
            table.insert(profile, user_hash_key)
            table.insert(profile, current_active_profile)
            table.insert(profile, redis.call('HGETALL', KEYS[1]))

            return profile                
        '''.format(local=int(local))

    @classmethod
    def delete_in_cache(cls):
        '''
        KEYS[1]: profile_key
        KEYS[2]: all profile keys key
        '''

        return '''
            redis.call('ZREM', KEYS[2], KEYS[1])
            redis.call('DEL', KEYS[1])
        '''

    @classmethod
    def get_all_user_pointers_if_main_profile(cls):
        '''
        KEYS[1]: user pointer set key
        ARGV[1]: user id
        '''

        return '''
                local all_user_pointers = redis.call('SMEMBERS', KEYS[1])
                local result = {}
                for idx, el in pairs(all_user_pointers) do
                    local pointed_user_id = redis.call('GET', el)
                    if pointed_user_id == nil or pointed_user_id~=ARGV[1] then
                        redis.call('SREM', KEYS[1], el)
                    else
                        table.insert(result, el)
                    end
                end
                return result
            '''


from setuptools import setup, find_packages
from pathlib import Path
import re
import os

PACKAGE_DIR = os.path.join(Path(__file__).resolve().parent, 'python_grains')
with open(os.path.join(PACKAGE_DIR, 'version.py')) as version_file:
    verstrline = version_file.read().strip()

VSRE = r'^__version__ = [\']([0-9\.]*)[\']'
mo = re.search(VSRE, verstrline, re.M)
if mo:
    VERSION = mo.group(1)
else:
    raise Exception('No version string found')


with open('README.md') as readme_file:
    README = readme_file.read()

setup_args = dict(
    name='python_grains',
    version=VERSION,
    description='Grains of Python',
    license='MIT',
    packages=find_packages(),
    author='Jacob Noordmans',
    author_email='jacob@graindataconsultants.com',
    keywords=['Grain', 'Python']
)

install_requires = [
    'redis',
    'pytz',
    'requests'
]

if __name__ == '__main__':
    setup(**setup_args, install_requires=install_requires)


import datetime
import pytz
import time
import random
import json
import uuid
import functools
import sys
import inspect

from python_grains.profiles import profile_properties
from python_grains.profiles import exceptions
from python_grains.locks import Lock

DEFAULT_DOMAIN = None
DEFAULT_REDIS_CLIENT = None
DEFAULT_DJANGO_PROFILE_MODEL = None
DEFAULT_DJANGO_COUNT_FUNC = None
DEFAULT_DJANGO_POINTER_MODEL = None
DEFAULT_DJANGO_CONNECTION_TUPLE = (None, None)
DEFAULT_PROFILE_TTL = 12 * 60 * 60
DEFAULT_PROFILE_TTL_MARGIN = 2 * 60 * 60
DEFAULT_USER_TTL = 6 * 24 * 60 * 60
DEFAULT_USER_DATA_TTL_MARGIN = 2
DEFAULT_FAST_CLEAN_THRESHOLD = 3
DEFAULT_CLEAN_QUEUE_TTL = 24 * 60 * 60

def set_data_type_properties(cls):

    cls.data_types = {
        prop.py_key: (prop.js_key, prop) for prop in
        [obj for name, obj in inspect.getmembers(sys.modules[profile_properties.__name__])
         if inspect.isclass(obj) and
         issubclass(obj, profile_properties.UsableProperty) and
         obj.__name__ != 'UsableProperty']
    }
    cls.property_by_prefix = {p[1].prefix: (k, p[1]) for k, p in cls.data_types.items()}

    for data_type in cls.data_types:
        setattr(cls, data_type, property(functools.partial(cls.get_property_dict, data_type=data_type)))
    return cls


@set_data_type_properties
class Profile(object):

    db_lookup_fail_check_ttl = 24 * 60 * 60
    domain = DEFAULT_DOMAIN
    fast_clean_threshold = DEFAULT_FAST_CLEAN_THRESHOLD
    redis_client = DEFAULT_REDIS_CLIENT
    django_profile_model = DEFAULT_DJANGO_PROFILE_MODEL
    django_pointer_model = DEFAULT_DJANGO_POINTER_MODEL
    django_connection_tuple = DEFAULT_DJANGO_CONNECTION_TUPLE
    django_count_func = DEFAULT_DJANGO_COUNT_FUNC
    profile_ttl = DEFAULT_PROFILE_TTL
    profile_ttl_margin = DEFAULT_PROFILE_TTL_MARGIN
    user_ttl = DEFAULT_USER_TTL
    user_data_ttl_margin = DEFAULT_USER_DATA_TTL_MARGIN
    clean_queue_ttl = DEFAULT_CLEAN_QUEUE_TTL

    def __init__(self,
                 key=None,
                 user_id=None,
                 profile_id=None,
                 session_id=None,
                 time=None,
                 raw_input=None,
                 updated=None,
                 created=None,
                 props_updated=None,
                 props_created=None,
                 init=True):

        now = pytz.utc.localize(datetime.datetime.utcnow())

        self.raw_input = raw_input
        self.user_id = user_id
        self.key = key
        self.session_id = session_id
        self.updated = updated or time or now
        self.created = created or time or now
        self.time = time or now
        self.profile_id = profile_id

        if self.domain is None:
            raise ValueError('domain should not be None')

        if self.redis_client is None:
            raise ValueError('redis_client should not be None')

        if init:
            self.build_properties(updated=props_updated,
                                  created=props_created)

    @classmethod
    def usable_django_model(cls):
        if cls.django_profile_model:
            if all(cls.django_connection_tuple):
                if cls.django_connection_tuple[0].connection and not cls.django_connection_tuple[0].is_usable():
                    del cls.django_connection_tuple[1]._connections.default
                return cls.django_profile_model
            else:
                return cls.django_profile_model

    def build_properties(self,
                         updated=None,
                         created=None):

        self.properties = {}

        for key, (data_key, data_type) in self.data_types.items():
            self.properties[key] = self.build_properties_for_type(data_key,
                                                                  data_type,
                                                                  updated=updated,
                                                                  created=created)

    def build_properties_for_type(self,
                                  data_key,
                                  data_type,
                                  updated=None,
                                  created=None):

        data = self.raw_input.get(data_key)

        if not data or len(data) == 0:
            return {}

        return {
            obj['name']: data_type(name=obj['name'],
                                   value=obj.get('value'),
                                   ttl=obj.get('ttl'),
                                   updated=updated,
                                   created=created)
            for obj in data
        }

    def __add__(self, other):

        if isinstance(other, int) and other == 0:
            return self

        assert self.__class__ == other.__class__, 'Cannot add {} to {}'\
            .format(self.__class__.__name__, other.__class__.__name__)

        assert self.domain == other.domain, 'Cannot add profiles for different domains'
        assert self.redis_id == other.redis_id, 'Cannot add profiles for different Redis instances'

        props1 = self.properties.copy()
        props2 = other.properties.copy()
        props3 = {}

        data_types = set(props1.keys()).union(set(props2.keys()))

        for data_type in data_types:
            p1 = props1.get(data_type, {})
            p2 = props2.get(data_type, {})
            p3 = {}
            names = set(p1.keys()).union(set(p2.keys()))
            for name in names:
                p3[name] = p1.get(name, 0) + p2.get(name, 0)
            props3[data_type] = p3

        updated = max(self.updated, other.updated)
        created = min(self.created, other.created)

        profile = self.__class__(user_id=self.user_id,
                                 raw_input=None,
                                 updated=updated,
                                 created=created,
                                 init=False)

        profile.properties = props3

        return profile

    def __radd__(self, other):

        return self.__add__(other)

    @classmethod
    def get(cls,
            user_id,
            time,
            force_db_lookup=False,
            fallback_to_db_lookup=False):

        # fallback_to_db_lookup: set to True to fallback to db IF no user profile is found at all in Redis
        # force_db_lookup: set to True do do a db search, no matter what.

        data = cls.redis_client.get_user_profile(
            keys=[
                cls._profile_pool_key(user_id)
            ],
            args=[
                time.timestamp()
            ]
        )
        # dictionary with keys: profile_id and values profile
        data_dict = dict(zip(data[::2], data[1::2]))
        ghost_keys = [key for key, data in data_dict.items() if not data]

        db_fallback = ((len(data_dict) == 0 and fallback_to_db_lookup) or len(ghost_keys) > 0) and \
                      not cls.db_lookup_failed_before(user_id=user_id)

        profiles_from_db = {}

        if force_db_lookup or db_fallback:

            profiles_from_db = cls.from_database_by_user_id(user_id)

            pipe = cls.redis_client.pipeline()
            for p in profiles_from_db.values():
                p.to_cache(key=p.profile_key, add_to_pool=True, pipe=pipe)

            pipe.delete(cls._new_user_detected_key(user_id=user_id))
            pipe.execute()

            data_dict = {k: v for k, v in data_dict.items()
                         if not cls.profile_id_from_key(k.decode()) in profiles_from_db}

        if db_fallback and len(profiles_from_db) == 0:
            cls.set_db_lookup_fail(user_id=user_id)

        profiles = [profile for profile in
                    [cls.from_redis_data(data=d,
                                         key=key,
                                         user_id=user_id,
                                         try_db=False) for key, d in data_dict.items()
                     if not d is None]
                    if not profile is None]

        profiles += list(profiles_from_db.values())

        if len(profiles) == 0:
            return None
        if len(profiles_from_db) >= cls.fast_clean_threshold:
            cls.user_in_cleaning_queue(user_id)

        return sum(profiles)

    @classmethod
    def from_database_by_user_id(cls,
                                 user_id):

        profile_models = cls.usable_django_model().objects.filter(user_id=user_id)
        data = {profile_model.profile_id: cls.from_flat_data(data=json.loads(profile_model.data),
                                                             user_id=user_id)
                for profile_model in profile_models}
        for profile_id, profile in data.items():
            profile.profile_id = profile_id
        return data

    @classmethod
    def db_lookup_failed_before(cls, user_id):

        return cls.redis_client.get(cls._user_not_found_in_db_key(user_id=user_id)) == b'1'

    @classmethod
    def set_db_lookup_fail(cls, user_id):

        cls.redis_client.set(cls._user_not_found_in_db_key(user_id=user_id), value='1',
                                        nx=True, ex=cls.db_lookup_fail_check_ttl)

    def store_in_db(self, data=None):

        max_tries = 3

        if self.key is None:
            self.key = self.new_profile_key

        if data is None:
            datastring = json.dumps(self.flat_data)
        else:
            datastring = json.dumps(data)

        for i in range(max_tries):
            try:
                self.usable_django_model().objects.update_or_create(
                    profile_id=self.profile_id_from_key(self.key),
                    user_id=self.user_id,
                    defaults={'data': datastring})
                return
            except Exception as e:
                if 'OperationalError' in e.__class__.__name__:
                    error = e
                    time.sleep(1 + random.random() / 10)
            raise error

    def delete_in_db(self):

        try:
            self.usable_django_model().objects.get(profile_id=self.profile_id).delete()
        except self.django_profile_model.DoesNotExist:
            pass

    def delete_in_cache(self):

        self.redis_client.delete_in_cache(
            keys=[
                self.key,
                self.all_profile_keys_key],
            args=[])

    @classmethod
    def switch_profiles(cls,
                        user_id,
                        new_profile_key,
                        old_profile_keys):

        return cls.redis_client.switch_profiles(
            keys=[
                cls._profile_pool_key(user_id=user_id),
                cls._all_profile_keys_key()
            ],
            args=[
                json.dumps(old_profile_keys),
                new_profile_key
            ]
        )

    def store_user_pointers(self):

        pointers = self.redis_client.get_all_user_pointers_if_main_profile(
            keys=[self.all_pointers_key],
            args=[self.user_id])

        pointers = [p.decode() if isinstance(p, bytes) else p for p in pointers]
        pointers = [p[len(self.pointer_prefix()):] for p in pointers]

        max_tries = 3
        error = None
        errors = []
        for pointer in pointers:
            for i in range(max_tries):
                try:
                    self.django_pointer_model.objects.update_or_create(
                        pointer=pointer,
                        defaults={'user_id': self.user_id}
                    )
                    error = None
                    break
                except Exception as e:
                    if 'OperationalError' in e.__class__.__name__:
                        error = e
                        time.sleep(1.0 + random.random() / 10)
                    else:
                        raise e
            if not error is None:
                errors.append(error)

        if len(errors) > 0:
            raise errors[0]

        return len(pointers)

    def reset_user_not_found(self):

        self.redis_client.delete(self.user_not_found_in_db_key)

    def remove_from_to_be_cleaned_queue(self):

        self.redis_client.srem(self.users_to_be_cleaned_key, self.user_id)

    @classmethod
    def get_stitched(cls,
                     user_ids,
                     time,
                     fallback_to_db_lookup=False):

        profile_pool_keys = [cls._profile_pool_key(user_id) for user_id in user_ids]
        data = cls.redis_client.get_stitched_user_profile(keys=[], args=[json.dumps(profile_pool_keys), time.timestamp()])
        # a dictionary with profile_pool_key as key and corresponding (profile_key - profile) dict as value
        data_dict = {k.decode(): {pk.decode(): pr for pk, pr in dict(zip(v[::2], v[1::2])).items()} if v else None
                     for k, v in dict(zip(data[::2], data[1::2])).items()}
        incomplete_users = {cls._user_id_from_profile_pool_key(k): v for k, v in data_dict.items() if not v or any(not bool(x) for x in v.values())}
        complete_users = {u:p for u,p in {cls._user_id_from_profile_pool_key(k): v for k, v in data_dict.items()}.items() if not u in incomplete_users}

        complete_profiles = [p for p in [cls.from_redis_data(data=d,
                                                             key=key,
                                                             user_id=user_id) for user_id, user_data in complete_users.items() for key, d in user_data.items()] if not p is None]

        missing_users = []

        if fallback_to_db_lookup:

            from_db = cls.from_database_by_user_ids(incomplete_users.keys())
            complemented_profiles = []
            for user_id, profiles_for_user in incomplete_users.items():
                from_db_for_user = from_db.get(user_id, [])
                if len(profiles_for_user) == 0:
                    if len(from_db_for_user) ==0:
                        missing_users.append(user_id)
                    else:
                        complemented_profiles.extend([cls.from_redis_data(profile.data,
                                                                          key=profile.key,
                                                                          user_id=user_id) for profile in from_db_for_user])
                else:
                    for profile in from_db_for_user:
                        if not profile.key in profiles_for_user:
                            profiles_for_user[profile.key] = profile.data
                complemented_profiles.extend(cls.from_redis_data(data=d,
                                                                 key=key,
                                                                 user_id=user_id) for key,d in profiles_for_user.items())
        else:
            complemented_profiles = [cls.from_redis_data(data=d,
                                                         key=key,
                                                         user_id=user_id) for user_id, user_data in incomplete_users.items() for key, d in user_data.items()]

        complemented_profiles = [p for p in complemented_profiles if not p is None]

        final_profiles = complete_profiles + complemented_profiles

        if len(final_profiles) == 0:
            return missing_users, None

        profile = sum(final_profiles)

        return missing_users, profile

        # if fallback is true, retrieve from database for those users (if users are missing in db, something is wrong,
        # however we should still delete them from stitch pool)

    def to_cache(self, key, add_to_pool=False, pipe=None):

        if pipe:
            pipe.hmset(key, self.flat_data)
            pipe.expire(key, self.profile_ttl + self.profile_ttl_margin)
            if add_to_pool:
                pipe.sadd(self.profile_pool_key, key)
                pipe.expire(self.profile_pool_key, self.user_ttl + self.user_data_ttl_margin)
        else:
            self.redis_client.hmset(key, self.flat_data)
            self.redis_client.expire(key, self.profile_ttl + self.profile_ttl_margin)
            if add_to_pool:
                self.redis_client.sadd(self.profile_pool_key, key)
                self.redis_client.expire(self.profile_pool_key, self.user_ttl + self.user_data_ttl_margin)

    def init_profile_to_cache(self,
                              time):

        created = self.redis_client.set_profile_if_new_user(
            keys=[
                self.user_hash_key,
                self.all_profile_keys_key,
                self.new_profile_key,
                self.profile_pool_key,
                self.new_user_detected_key
            ],
            args=[
                json.dumps([y for x in self.flat_data.items() for y in x]),
                time.timestamp()
            ]
        )
        if created is None:
            return False
        return bool(int(created.decode()))

    @classmethod
    def from_database_by_user_ids(cls,
                                  user_ids):

        user_ids = [u for u in user_ids if not cls.db_lookup_failed_before(u)]
        profile_models = cls.usable_django_model().objects.filter(user_id__in=user_ids)

        result = {}

        for profile_model in profile_models:
            user_id = str(profile_model.user_id)
            profile = cls.from_flat_data(data=json.loads(profile_model.data),
                                         user_id=user_id)
            profile.profile_id = profile_model.profile_id
            result[user_id] = result.get(user_id, []) + [profile]

        failed_users = [u for u in user_ids if not u in result]
        for failed_user_id in failed_users:
            cls.set_db_lookup_fail(failed_user_id)

        return result

    @classmethod
    def from_database(cls,
                      key,
                      user_id):
        try:
            profile_model = cls.usable_django_model().objects.get(profile_id=cls.profile_id_from_key(key),
                                                             user_id=user_id)
        except cls.django_profile_model.DoesNotExist:
            return None
        return cls.from_flat_data(data=json.loads(profile_model.data),
                                  user_id=user_id)

    @classmethod
    def from_flat_data(cls,
                       data,
                       user_id):

        names = {'_'.join(n.split('_')[:-1]) for n in data if not n in ('_updated', '_created', '_user_hash')}

        properties = {}

        for data_type_name in names:

            _t = data_type_name.split('_')
            prefix, name = _t[0] + '_', '_'.join(_t[1:])
            if not prefix in cls.property_by_prefix:
                continue
            data_type, profile_property = cls.property_by_prefix[prefix]
            props = properties.get(data_type, {})
            value = data[data_type_name + profile_properties.PROFILE_VALUE_POSTFIX]
            updated = cls.timestamp_to_datetime(data.get(
                data_type_name + profile_properties.PROFILE_VALUE_UPDATED_POSTFIX))
            created = cls.timestamp_to_datetime(data.get(
                data_type_name + profile_properties.PROFILE_VALUE_CREATED_POSTFIX))
            expire_at = cls.timestamp_to_datetime(
                data.get(data_type_name + profile_properties.PROFILE_VALUE_EXPIRE_AT_POSTFIX))
            props[name] = profile_property.from_redis_data(name=name,
                                                           value=value,
                                                           updated=updated,
                                                           created=created,
                                                           expire_at=expire_at)
            properties[data_type] = props

        _updated = cls.timestamp_to_datetime(data['_updated'])
        _created = cls.timestamp_to_datetime(data['_created'])

        profile = cls(user_id=user_id,
                      updated=_updated,
                      created=_created,
                      init=False)

        profile.properties = properties

        return profile

    @classmethod
    def get_to_be_cleaned_users(cls,
                                number=1000):
        user_ids = cls.redis_client.srandmember(cls._users_to_be_cleaned_key(), number=number)
        return [user_id.decode() if isinstance(user_id, bytes) else user_id for user_id in user_ids]

    @classmethod
    def get_expired(cls,
                    num=300,
                    start=0,
                    dev=False):

        now = pytz.utc.localize(datetime.datetime.utcnow()).timestamp()
        expired_profiles = cls.redis_client.zrangebyscore(cls._all_profile_keys_key(),
                                                          min='-inf',
                                                          max='+inf' if dev else (now - cls.profile_ttl),
                                                          start=start,
                                                          num=num,
                                                          withscores=True)
        return [cls(key=profile_key.decode(),
                    time=pytz.utc.localize(datetime.datetime.utcfromtimestamp(timestamp)),
                    init=False) for profile_key, timestamp in expired_profiles]

    def lock(self, ttl):
        return Lock(self.profile_id_from_key(self.key), self.redis_client, ttl=ttl * 1000)

    def load_and_finalize(self):

        data = self.redis_client.finalize_profile(keys=[self.key], args=[])

        if data is None:
            raise exceptions.ProfileNotFoundError

        user_hash_key, currently_active, profile_data = data
        self.user_id = self.user_id_from_user_hash_key(user_hash_key.decode())
        self.currently_active = currently_active == b'1'
        profile_data = dict(zip(profile_data[::2], profile_data[1::2]))
        self.profile_data = {k.decode(): v.decode() if isinstance(v, bytes) else v for k, v in profile_data.items()}

    def completely_remove_from_cache(self):

        self.redis_client.completely_remove_from_cache(
            keys=[
                self.profile_key,
                self.profile_pool_key,
                self.all_profile_keys_key
            ],
            args=[])

    def update(self,
               pointers=None,
               stitch_keys=None):

        new_user_detected = self.redis_client.update_profile(
            keys=[
                self.user_hash_key,
                self.all_profile_keys_key,
                self.all_pointers_key,
                self.new_profile_key,
                self.profile_pool_key,
                self.new_user_detected_key,
                self.stitch_pool_key],
            args=[
                self.lua_update_string,
                self.time.timestamp(),
                str(self.session_id or 'none'),
                str(self.user_id),
                json.dumps(self.create_pointers(pointers)),
                json.dumps(self.create_stitch_keys(stitch_keys))
            ]
        )
        return bool(int(new_user_detected.decode()))

    @classmethod
    def user_in_cleaning_queue(cls, user_id):

        cls.redis_client.sadd(cls._users_to_be_cleaned_key(), str(user_id))
        cls.redis_client.expire(cls._users_to_be_cleaned_key(), cls.clean_queue_ttl)

    @property
    def stitch_pool_key(self):

        return f'st:u:p:{self.domain}:{self.user_id}'

    def create_pointers(self, pointers):

        if pointers is None:
            return []
        return [self.pointer_from_input(p) for p in pointers]

    def create_stitch_keys(self, stitch_keys):

        if stitch_keys is None:
            return []
        return stitch_keys

    @property
    def lua_update_string(self):

        data = {}

        for data_type, properties in self.properties.items():
            redis_keys = [prop.redis_keys for prop in properties.values()]
            if redis_keys:
                data.update({data_type: redis_keys})
        return json.dumps(data)

    @staticmethod
    def get_property_dict(instance, data_type):

        return instance.properties.get(data_type, {})

    @staticmethod
    def to_short_data(props):

        return {k: v.output_value for k, v in props.items()}

    @property
    def data(self):

        data = {
            'start': self.created.isoformat(),
            'startTimestampMillis': self.created.timestamp() * 1000,
            'updated': self.updated.isoformat(),
            'updatedTimestampMillis': self.updated.timestamp() * 1000,
            **{v[0]: self.to_short_data(getattr(self, k)) for k, v in self.data_types.items()}
        }
        return {k: v for k, v in data.items() if isinstance(v, float) or len(v) > 0}

    @classmethod
    def timestamp_to_datetime(cls, value):

        if value is None:
            return None
        return pytz.utc.localize(datetime.datetime.utcfromtimestamp(float(value)))

    @classmethod
    def get_user_ids_with_multiple_profiles(cls,
                                            num,
                                            start):

        result = cls.usable_django_model().objects.all().values('user_id').annotate(
            count=cls.django_count_func('user_id')).filter(count__gt=1).order_by()[start:start + num]
        return [str(res['user_id']) for res in result]

    @classmethod
    def multiple_from_database(cls,
                               keys,
                               user_id):

        keys = [key.decode() if isinstance(key, bytes) else key for key in keys]
        profile_ids = [cls.profile_id_from_key(key) for key in keys]

        profile_models = list(cls.usable_django_model().objects.filter(profile_id__in=profile_ids, user_id=user_id))
        profiles = {cls._profile_key(profile_model.profile_id):
                        cls.from_flat_data(data=json.loads(profile_model.data),
                                           user_id=user_id)
                    for profile_model in profile_models}

        for profile_key, profile in profiles.items():
            profile.to_cache(key=profile_key)
        missing = list(set(profile_ids) - set(str(profile_model.profile_id) for profile_model in profile_models))
        for profile_id in missing:
            p = cls(user_id=user_id,
                    profile_id=profile_id,
                    init=False)
            p.completely_remove_from_cache()
        return profiles

    @classmethod
    def from_redis_data(cls,
                        data,
                        key,
                        user_id,
                        try_db=True):

        key = key.decode() if isinstance(key, bytes) else key
        if not data:
            if try_db:
                profile = cls.from_database(key,
                                            user_id=user_id)
                if not profile is None:
                    profile.to_cache(key)
                return profile
            return None
        else:
            data = [x.decode() if isinstance(x, bytes) else x for x in data]
            data = dict(zip(data[::2], data[1::2]))
            return cls.from_flat_data(data=data,
                                      user_id=user_id)

    @property
    def flat_data(self):

        data = {'_updated': self.updated.timestamp(),
                '_created': self.created.timestamp(),
                '_user_hash': self.user_hash_key}

        for data_type, properties in self.properties.items():
            for prop in properties.values():
                data.update(prop.flat_redis_keys)
        return data

    @property
    def user_hash_key(self):
        return self._user_hash_key(user_id=self.user_id)

    @classmethod
    def _user_hash_key(cls, user_id):
        return f'u:h:{cls.domain}:{user_id}'

    @classmethod
    def user_id_from_user_hash_key(cls, user_hash_key):
        return user_hash_key.split(':')[-1]

    @staticmethod
    def profile_id_from_key(key):

        return key.split(':')[-1]

    @property
    def all_profile_keys_key(self):

        return self._all_profile_keys_key()

    @classmethod
    def _all_profile_keys_key(cls):

        return f'u:pk:{cls.domain}:all'

    @property
    def profile_pool_key(self):

        return self._profile_pool_key(user_id=self.user_id)

    @classmethod
    def _profile_pool_key(cls, user_id):

        return f'u:pp:{cls.domain}:{user_id}'

    @classmethod
    def _user_id_from_profile_pool_key(cls, profile_pool_key):

        return profile_pool_key.split(':')[-1]

    @property
    def new_user_detected_key(self):

        return self._new_user_detected_key(user_id=self.user_id)

    @classmethod
    def _new_user_detected_key(cls, user_id):

        return f'u:nd:{cls.domain}:{user_id}'

    @property
    def user_not_found_in_db_key(self):
        return self._user_not_found_in_db_key(user_id=self.user_id)

    @classmethod
    def _user_not_found_in_db_key(cls, user_id):
        return f'u:nf:{cls.domain}:{user_id}'

    @property
    def new_profile_key(self):

        self.profile_id = str(uuid.uuid4())
        return self.profile_key

    @property
    def users_to_be_cleaned_key(self):
        return self._users_to_be_cleaned_key()

    @classmethod
    def _users_to_be_cleaned_key(cls):
        return f'u:tbc:{cls.domain}'

    @property
    def profile_key(self):
        return self._profile_key(profile_id=self.profile_id)

    @classmethod
    def _profile_key(cls, profile_id):
        return f'u:p:{cls.domain}:{profile_id}'

    @property
    def all_pointers_key(self):
        return self._all_pointers_key(user_id=self.user_id)

    @classmethod
    def _all_pointers_key(cls, user_id):
        return f'u:pt:{cls.domain}:{user_id}'

    @staticmethod
    def pointer_from_input(value):
        return f'{Profile.pointer_prefix()}{value}'

    @staticmethod
    def pointer_prefix():
        return 'u:pt:'

    @property
    def redis_id(self):
        if not hasattr(self, '_redis_id') or self._redis_id is None:
            conn_kwargs = self.redis_client.get_connection_kwargs()
            self._redis_id = f'{conn_kwargs["host"]}:{conn_kwargs["port"]}:{conn_kwargs["db"]}'
        return self._redis_id